import os
import time
import cpuinfo
import nvsmi
import psutil
import numpy as np
import pandas as pd
from PIL import Image, ImageOps
import dorothy_sdk.dataset_wrapper as dw
import dorothy_sdk.mlflow as dmlf
from lpsds.inout import save, get_path

########### INSTRUÇÕES DE DESENVOLVIMENTO ############################
#
# Muitas das funções abaixo já foram implementadas de forma que não precisarão ser alteradas. Muito, MUITO provavelmente
# você só precisará alterar para os eu caso específico, o método "operate_model" (veja em operation_specific.py um exemplo)
#
# Caso vc REALMENTE precise mexer em alguns dos métodos já implementdos abaixo, faça através de uma nova classe que herde
# de ModelOperation, mexendo somente nos métodos necessários e respeitando sempre os requisitos de entrada e saída de
# cada método.
#

class ModelOperation:
    """
    Esta função define a os métodos necessários para a operação do seu modelo.

    Muito provavelmente você só precisará implementar o método operate_model.
    """

    def __init__(self, mlflow_module_flavor, device_type='cpu',
                    img_path_col_name='image_local_path', label_col_name='y_true'):
        """
        Construtor da classe.
        
        Parâmetros:
        - mlflow_module_flavor: flavor do mlflow do seu modelo. Opções
                                (entre outras), são mlflow.pytorch, mlflow.sklearn, mlflow.tensorflow.
        - device_type: se a execução do modelo se dará em 'cpu' ou 'gpu'.
        - img_path_col_name: nome da coluna no dataframe de metadados que contém o path local para as imagens.
        - label_col_name: nome da coluna no dataframe que contém os alvos do modelo (1 ou 0)
        """
        assert device_type in ['gpu', 'cpu'], 'device_type must be either "cpu" or "gpu"'
        self.mlflow_module_flavor = mlflow_module_flavor
        self.runs_path = 'output/runs'
        self.models_path = 'output/models'
        self.device_type = device_type
        self.img_path_col_name = img_path_col_name
        self.label_col_name = label_col_name
        self.device = self.get_device()
    

    def operate_model(self, model, X, decision_threshold: float=0.5):
        """
        Aqui você deve definir o código responsável por passar o conjunto de imagens pelo modelo.
        Esta função deve retornar 2 numpy.arrays:
            - y_proba: contendo o valor da saída do modelo para cada amostra em X.
            - y_pred: contendo a decisão do modelo para cada amostra em X

        VOCÊ DEVE ALTERAR ESTA FUNÇÃO PARA O SEU CASO ESPECÍFICO, ATENTANDO QUE ELA
        PRECISA MANTER INALTERADOS SEUS PARÂMETROS DE ENTRADA E SAÍDA.
        """
        raise NotImplemented


    def predict(self, model, X, decision_threshold: float=0.5) -> float:
        """
        Esta função deve prover a decisão do modelo para uma única imagem fornecida
        (não pode ser um conjunto de imagens).
        """
        raise NotImplemented


    def train_model(self, trn_df: pd.DataFrame , val_df: pd.DataFrame, **kwargs):
        """
        Esta função é responsável por treinar um modelo (sem CV) a partir de um dataframe de treino e um de validação.
        O modelo retornado deve ser capaz de ser passado para os métodos operate_model e predict.

        Input:
          - trn_df: dataframe contendo os dados de treinamento (path para as imagens e seus alvos)
          - val_df: dataframe contendo os dados de validação (path para as imagens e seus alvos)
          - **kwargs: qualquer hiperparâmetro necessário ao treinamento do modelo.

        Return: modelo treinado.
        """
        raise NotImplemented
    

    def _save_model(self, model, file_name: str) -> None:
        """
        Esta função deve definir a lógica necessária para salvar um modelo no path passado.
        Entrada:
          - model: uma instância de modelo treinada.
          - file_name: path onde salvar o modelo.
        """
        raise NotImplemented


    def _load_model(self, file_name: str):
        """
        Esta função deve definir a lógica necessária para carregar um modelo salvo em disco.
        O modelo retornado deve ser capaz de ser passado para os métodos operate_model e predict.

        Entrada:
          - file_name: path onde salvar o modelo.
        
        Retorno: a instância treinada salva no local especificado.
        """
        raise NotImplemented


    def get_device(self):
        """
        Deve retornar o identificador do dispositivo a ser usado ('/CPU:0', torch.device('cuda'), etc.)
        de acordo com o valor de self.device_type, que pode ser cpu ou gpu.
        """
        raise NotImplemented


    def get_dataset_images(self, validation_dataset_name):
        """
        Nesta função, vc deve colocar o código necessário para carregar as imagens de um dataset do DORTHY
        Esta função deve retornar:

            - um tensor X que possa ser passado diretamente ao método "predict" do modelo.
            - um pandas.DataFrame com os metadados deste dataset (id de cada imagem, sexo, idade, etc.)
        
        MUITO PROVAVELMENTE VOCÊ NÃO PRECISARÁ MEXER NESSA FUNÇÃO
        """
        dataset = dw.Dataset(validation_dataset_name)
        metadata = dataset.load_metadata()
        X,_ = dw.Dataset.load_from_file(metadata)
        return X, metadata


    def get_operation_model(self, mlflow_run_id):
        """
        Nesta função, vc deve implementar o código necessário para buscar no MLFlow server o SEU MODELO DE OEPRAÇÃO.
        A saída desta função deve ser um modelo pronto para ser operado através das suas funções predict e predict_proba.

        O MLFlow tem "sabores" específicos para vários frameworks (PyTorch, tensorflow, sklearn, etc.). Você deve usar
        o sabor correto (no exemplo abaixo, o modelo foi produzido pelo pytorch, logo o sabor correto aqui é o mlflow.pytorch).
        Veja detalhes aqui: https://mlflow.org/docs/1.8.0/python_api/index.html

        MUITO PROVAVELMENTE VOCÊ NÃO PRECISARÁ MEXER NESSA FUNÇÃO
        """
        run = dmlf.MLFlow(mlflow_run_id, flavor=self.mlflow_module_flavor)
        model = run.load_model()
        threshold = float(run.get_params().decision_threshold)
        return model, threshold
    

    def get_processing_stats(self, model, X) -> pd.DataFrame:
        """
        Esta função gera estatística de execução tais como tempo de processamento, memória consumida, etc.
        """

        #Pre-allocating the values storage space so it won´t
        #interfere with the memory usage calculation.
        elapsed_time = np.empty(X.shape[0])
        memory_used = np.empty(X.shape[0])
        process = psutil.Process(os.getpid())

        for i,x in enumerate(X):
            start_time = time.time()
            self.predict(model, x)
            elapsed_time[i] = time.time() - start_time
            memory_used[i] = process.memory_info().rss

        cpu = cpuinfo.get_cpu_info()
        gpu = tuple(nvsmi.get_gpus())[0]
        max_allowed_cpu = os.environ.get('NUM_ALLOWED_CPUS', None)
        #NUM_ALLOWED_CPUS may come as NUM_ALLOWED_CPUS=''
        if not max_allowed_cpu: max_allowed_cpu = cpu['count']

        ret = pd.DataFrame({'elapsed_time_sec' : elapsed_time, 'memory_used_bytes' : memory_used})
        ret['num_logical_cpu'] = cpu['count']
        ret['max_logical_cpu_allowed'] = max_allowed_cpu
        ret['cpu_type'] = cpu['brand_raw']
        ret['cpu_clock_mhz'] = psutil.cpu_freq(percpu=False).current
        ret['cpu_mem_gb'] = psutil.virtual_memory().total / (1024**3)
        ret['gpu_type'] = gpu.name
        ret['gpu_mem_gb'] = gpu.mem_total / 1024
        ret['device_used'] = self.device_type


        return ret


    def save_results(self, path, dataset_metadata, y_proba=None, y_pred=None):
        """
        Aqui os vetores de saídas são incorporados ao dataframe de metadados proveniente do DOROTHY e
        salvos na pasta correta para análise posterior.

        MUITO PROVAVELMENTE VOCÊ NÃO PRECISARÁ MEXER NESSA FUNÇÃO
        """

        if y_proba is not None: dataset_metadata['y_proba'] = y_proba
        if y_pred is not None: dataset_metadata['y_pred'] = y_pred
        dataset_metadata.to_parquet(path)


    def save_model(self, model, path, threshold :float=0.5):
        """
        Salva o modelo produzido.

        MUITO PROVAVELMENTE VOCÊ NÃO PRECISARÁ MEXER NESSA FUNÇÃO
        """
        self._save_model(model, path, threshold)
    

    def load_model(self, fname :str):
        """
        carrega um modelo salvo em disco. Precisa retornar uma tuple (model, threshold)
        """
        full_path = get_path(dir_name=self.models_path, fname=fname)
        return self._load_model(full_path)


def get_fold_data(cv_model: dict, cv_splits: list, fold_idx: int) -> tuple:
     """"
     def get_fold_data(cv_model: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], fold_idx: int) -> tuple
     Returns the data corresponding to a given cross validation fold.
     Input parameters:
       - cv_model: a dictionary with structure equivalent to the one returned by sklearn.model_selection.cross_validate.
       - cv_splits: a list where each item is a tuple containing the indexes to be used for training and testing in
                    see dataset_wrapper.Dataset.train_test_split, for isntance, for details.
       - fold_idx: the index of the cv fold you want to collect the data.
    
     Returns:
       - tst_df: the target set used for testing in the desired fold as a dataframe.
       - model: the operation model yielded by the desired fold.
     """
     _, tst_df = cv_splits[fold_idx]
     model = cv_model['estimator'][fold_idx]
    
     return tst_df, model



def get_operation_model(cv_model: dict, cv_splits: list, metric:str='test_sp', less_is_better:bool=False) -> tuple:
    """"
    get_operation_model(cv_model: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], metric:str='test_sp', less_is_better:bool=False) -> tuple
    Returns the operation model and its corresponding data. For that, it finds
    the model yielding the best metric over its corresponding testing set.
    Input parameters:
      - cv_model: a dictionary with structure equivalent to the one returned by sklearn.model_selection.cross_validate.
      - cv_splits: a list where each item is a tuple containing the indexes to be used for training and testing in
                   see sklearn.model_selection.KFold, for isntance, for details.
      - metric: a key within cv_model containing the metric to use to compare cv folds.
      - less_is_better: if True, the best fold will be the one where the passed metric is the lowest (MSE, for instance). The highest
                        will be returned if this value is set to True (Accuracy, for instance).
    
    Returns:
      - best_fold: the id of the fold yielding the operation model.
       - tst_df: the target set used for testing in the desired fold as a dataframe.
       - model: the operation model yielded by the desired fold.
    """
    best_fold = cv_model[metric].argmin() if less_is_better else cv_model[metric].argmax()
    tst_df, model = get_fold_data(cv_model, cv_splits, best_fold)
    return best_fold, tst_df, model


"""
Repository-related operations
"""
import os
import tempfile
import git

class Repository:
    """
    Abstracts git-related operations.
    """
    def __init__(self, path :str=os.getcwd(), version :str=None):
        """
        def __init__(self, path :str=os.getcwd(), version :str=None):

        Receives the path to the repository you wish to manage.

        Input:
          - path: a local or remote (starting with 'git@github.com') path to the repositpry you wish to manage.
          - version: a tag name you with to checkout.

        """
        if path.startswith('git@github.com'): #It is a remote repository
            self.remote_url = path
            self.path = tempfile.TemporaryDirectory()
            self.rep = git.Repo.clone_from(self.remote_url, self.path.name)
        else: #it is a local repository
            self.remote_url = None
            self.path = path
            self.rep = git.Repo(self.path)
        
        if version:
            self.rep.git.checkout(version)


    def get_version(self) -> str:
        """
        def get_version(self) -> str

        Returns the version of the cloned repository.
        """
        return self.rep.git.describe(tags=True)

    
    def get_location(self):
        """
        def get_version(self) -> str

        Returns the location (URL) of the cloned repository in GtHub.
        """
        return self.rep.remotes.origin.url



"""
Módulo de apoio ao uso do MLFlow
"""

import mlflow
import json
import lpsds.mlflow
from dorothy_sdk.dataset_wrapper import Dataset


class MLFlow(lpsds.mlflow.MLFlow):

    def __init__(self, run_id=None, flavor=mlflow.sklearn):
        """
        Class constructor.

        Input:
          - run_id. The MLFlow run ID you want to manipulate. If None, then a new run is created.
          - flavor: the model flavor (mlflow.sklearn, mlflow.tensoflow, etc.) to be use to manage trained models.
        """
        super().__init__(run_id)
        self.flavor = flavor


    def _get_model_path(self, partition :int=None, split :int=None) -> str:
        """
        Defines the path from where to collect the trained module.

        Input:
          - partition: the CV fold (0-9) from which the model was obtained. If None, then
                       saves as the operation model.
          - split: which split to index (0-8) to id the model. If None, then
                       saves as the operation model.
        
          Only one value set to None suffices to return the operation model path.
        Returns: the folder name where the model can be found.
        """
        if partition is None or split is None:
          return 'operation_model'
        return f'model_part_{partition}_split_{split}'


    def load_model(self, partition :int=None, split :int=None, **kwargs):
        """
        Gets the operating model or the model from a specific fold.

        Input:
          - partition: the CV fold (0-9) from which the model was obtained. If None, then
                       saves as the operation model.
          - split: which split to index (0-8) to id the model. If None, then
                       saves as the operation model.
          - **kwargs: passed to mlflow.load_model
        
          Only one value set to None suffices to return the operation model path.          

        Returns: an instance to the trained object.
        """
        fold = self._get_model_path(partition, split)
        return self.flavor.load_model(self.run.info.artifact_uri + '/' + fold, **kwargs)
  

    def log_model(self, model, partition :int=None, split :int=None, **kwargs):
        """

        Saves the operating model or the model from a specific fold.

        Input parameters:
          - model: the trained module to be saved.
          - partition: the CV fold (0-9) from which the model was obtained. If None, then
                       saves as the operation model.
          - split: which split to index (0-8) to id the model. If None, then
                       saves as the operation model.
          - **kwargs: passed to mlflow.log_model
        
        Only one value set to None suffices to return the operation model path.
        """
        fold = self._get_model_path(partition, split)
        self.flavor.log_model(model, fold, registered_model_name="ImageClassifier", **kwargs)


    def load_dataset(self, dataset_name, perform_cv_split=False):
      dataset_list = json.loads(self.run.data.params[dataset_name])
      return Dataset.join_datasets(dataset_list, perform_cv_split=perform_cv_split)



"""
This module acts as a wrapper on top of DOROTHY sdk low level features.

It creates a cached environment and save files in a format more easily read
than reading image by image.
"""

import os
import pickle
import numpy as np
import pandas as pd
import dorothy_sdk as sdk
from PIL import Image, ImageOps
from typing import List, Tuple


class Dataset:
    """
    Base class.
    """
    def __init__(self, dataset):
        self.dataset = dataset
        self.client = sdk.Client()
        self.data_file_name = f'{self.dataset}-data.npy'
        self.metadata_file_name = f'{self.dataset}-metadata.parquet'
        self.cv_split_file_name = f'{self.dataset}-cv_split.parquet'
        self.cache_location = self._get_cache_location()
        self.image_width = 256
        self.image_height = 256
        self.image_grayscale = True

    def _get_cache_location(self):
        """Gets the cache location from dorothy sdk module."""
        # img_sample = self.client.dataset(self.dataset).list_images()[0]
        # return img_sample._cache_location
        return os.environ['DOROTYSDK_CACHE_LOCATION']

    def _cache_exists(self, data_name):
        return os.path.exists(os.path.join(self.cache_location, data_name))

    def load_metadata(self):
        """
        Returns the dataset metadata, which includes:
          - Image file names
          - Image target (y_true)
          - CV split info
        """
        if not self._cache_exists(self.metadata_file_name):
            self._assemble_dataset()
        return pd.read_parquet(os.path.join(self.cache_location, self.metadata_file_name))


    def load_cv_splits(self):
        """Returns a list of trn and tst indexes for each fold"""
        if not self._cache_exists(self.cv_split_file_name):
            self._assemble_cv_splits()
        return pd.read_parquet(os.path.join(self.cache_location, self.cv_split_file_name))

    
    def _assemble_dataset(self):
        """Download and create a local version of the dataset"""
        ds = self.client.dataset(self.dataset)
        metadata_list = []
        for img in ds.list_images():
            img = img.download_image(width=self.image_width, height=self.image_height, gray_scale=self.image_grayscale)
            image_local_path = os.path.join(self.cache_location, img._cached_image_name) if img._cache_enabled else None
            metadata_list.append({'image_url' : img.image_url, 'metadata' : img.metadata, 'image_local_path' : image_local_path})

        #Creating the metadata dataframe and saving it.
        df = self._assemble_metadata(metadata_list)
        df['y_true'] = 1
        df.loc[df.has_tb == False, 'y_true'] = 0
        meta_file_name = os.path.join(self.cache_location, self.metadata_file_name)
        df.to_parquet(meta_file_name)

    
    def _assemble_metadata(self, metadata_list):
        df = pd.DataFrame(columns=['url', 'image', 'gender', 'age', 'synthetic', 'date_exam', 'original_report', 'has_tb', 'additional_information', 'image_local_path'])
        for id, img in enumerate(metadata_list):
            image_name = img['image_url'][:-1] if img['image_url'][-1] == '/' else img['image_url']
            image_name = os.path.split(image_name)[1]
            df.loc[len(df)] = (img['image_url'],
                                image_name,
                                img['metadata']['gender'],
                                img['metadata']['age'],
                                img['metadata']['synthetic'],
                                img['metadata']['date_exam'],
                                img['metadata']['original_report'],
                                img['metadata']['has_tb'],
                                img['metadata']['additional_information'],
                                img['image_local_path'],
                                )

        #Saving metadata
        df.gender.replace(to_replace={'F' : 'female', 'M' : 'male'}, inplace=True)
        return df


    def _assemble_cv_splits(self):
        cluster = self.client.cluster(f"cluster_{self.dataset}")
        file = cluster.download_file()
        partition = pickle.loads(file)

        cols = ['partition', 'inner_split', 'set_type', 'image']
        df = pd.DataFrame(columns=cols)

        for i1, d1 in enumerate(partition):
            for i2, d2 in enumerate(d1):
                for i3, d3 in enumerate(d2):
                    for img_name in d3:
                        df.loc[len(df)] = i1, i2, i3, img_name
        
        set_map = {0 : 'train', 1 : 'validation', 2 : 'test'}
        df['set_type'] = df.set_type.map(set_map).astype(str)
        file_name = os.path.join(self.cache_location, self.cv_split_file_name)
        df.to_parquet(file_name)


    @staticmethod    
    def join_datasets(datasets, perform_cv_split=True) -> pd.DataFrame:
        """
        def join_datasets(datasets):
        
        Concatena os cv_splits de cada dataset especificado, e os enriquece com os metadados,
        gerando uma tabela completa com todas as infos disponíveis para o conjunto de datasets.

        Entrada: múltiplos nomes de datasets. i.e: join_datasets(['china', 'manaus']).

        Retorno: um dataframe com os cv splits enriquecidos com os metadados de cada imagem.
        """
        
        ds_list = []
        for ds_name in datasets:
            ds = Dataset(ds_name)
            meta = ds.load_metadata()
            meta['dataset'] = ds.dataset
            
            if perform_cv_split:
                cv_splitter = ds.load_cv_splits()
                meta = cv_splitter.merge(meta, how='inner', on='image')
            
            ds_list.append(meta)
        
        return pd.concat(ds_list, ignore_index=True)


    @staticmethod
    def train_test_split(cv_df :pd.DataFrame) -> List[ Tuple[pd.DataFrame, List[Tuple[pd.DataFrame, pd.DataFrame]]] ]:
        """    
        Cria uma lista de indices de validação cruzada no estilo do train_test_split do sklearn, porem usando
        a estrutura de CV do projeto. Isso significa que ela retorna todos as splits tst x treino onde treino
        é uma lista de splits trn x val.

        Entrada:
        - cv_df: um dataframe do tipo do retornado por ds.load_cv_splits()
        
        Retorno: uma lista de tuples onde cada tuple contém (inner_split_cv, tst_metadata), onde inner_split_cv
                é uma lista contendo tuples (trn_df, val_df).
        """
        
        ret = []
        for partition in cv_df.partition.drop_duplicates().sort_values():
            tst_df = cv_df.loc[(cv_df.partition==partition) & (cv_df.set_type == 'test')].drop_duplicates('image')
            
            inner_split_list = []
            for inner_split in cv_df.inner_split.drop_duplicates().sort_values():
                inner_df = cv_df.loc[(cv_df.partition==partition) & (cv_df.inner_split==inner_split)]
                val_df = inner_df.loc[inner_df.set_type == 'validation'].drop_duplicates('image')
                trn_df = inner_df.loc[inner_df.set_type == 'train'].drop_duplicates('image')
                inner_split_list.append( (trn_df, val_df) )

            ret.append( (inner_split_list, tst_df) )

        return ret

    @staticmethod
    def load_from_file(df: pd.DataFrame, path_col_name: str='image_local_path', label_col_name: str='y_true'):
        img_list = [Dataset.pre_process_image(Image.open(r[path_col_name])) for _,r in df.iterrows()]
        X = np.concatenate(img_list, axis=0)
        y_true = df[label_col_name].to_numpy()
        return X, y_true

    @staticmethod
    def pre_process_image(img :Image) -> np.ndarray:
        """Make a local image complient. You should use this method after
        loading an image either from DOROTHY or local path (cache)"""
        img = ImageOps.grayscale(img)
        #We must add a dimension here so concatenate will concatenate over it later.
        return np.expand_dims(np.array(img), axis=0)



import re
import warnings

from requests import Session

from dorothy_sdk.utils import url_join


class CrossValidationFolders:
    resource = "dataset/cross_validation/"
    dataset: str = None
    cluster_id: str = None
    file_url: str = None

    def __init__(self, session: Session, host: str, **kwargs):
        self._session: Session = session
        self._service_host = host
        if kwargs.get("dataset"):
            self.dataset = kwargs.get("dataset")

    def get(self, cluster_id: str = None, dataset: str = None, file_path: str = None):
        warnings.warn("Deprecated method 'get'", DeprecationWarning)
        return self.download(file_path)

    def download(self, file_path: str = None):
        url = url_join(self._service_host, self.resource, self.dataset)
        if url.endswith("/"):
            url = url[:-1]
        with self._session.get(url, stream=True) as request:
            request.raise_for_status()
            if file_path is None:
                file_path = re.findall("filename=(.+)", request.headers.get("Content-Disposition"))[0]
            with open(file_path, 'wb') as f:
                for chunk in request.iter_content(chunk_size=8192):
                    f.write(chunk)
        return file_path

    def download_file(self, path: str = None):
        warnings.warn("Deprecated method 'download_file'", DeprecationWarning)
        self.download(path)



from typing import List, Iterable
from requests import Session
from dorothy_sdk.utils import url_join
from dorothy_sdk.resources.image import Image
from dorothy_sdk.resources.folders import CrossValidationFolders


class Dataset:
    resource = "datasets"

    def __init__(self, name: str, image_formats: str, number_images: int, session: Session, host: str, *args, **kwargs):
        self.number_images = number_images
        self.name = name
        self.image_formats = image_formats
        self._session = session
        self._service_host = host

    def list_images(self, enable_pagination=False):
        if enable_pagination:
            request = self._session.get(url_join(self._service_host, 'dataset', self.name, 'images'))
            request.raise_for_status()
            if request.status_code == 200:
                response = request.json()
                while response.get("next", None) is not None:
                    for data in response.get("results"):
                        yield Image(session=self._session, host=self._service_host, **data)
                    request = self._session.get(response.get("next"))
                    response = request.json()
                if response.get("next", None) is None and len(response.get("results", [])) > 0:
                    for data in response.get("results"):
                        yield Image(session=self._session, host=self._service_host, **data)
            else:
                raise RuntimeError("Unable to list available images")
        else:
            request = self._session.get(url_join(self._service_host, Image.resource), params={"search": self.name})
            request.raise_for_status()
            if request.status_code == 200:
                for element in request.json():
                    yield Image(session=self._session, host=self._service_host, **element)
            else:
                raise RuntimeError("Unable to list available images")

    def get_image(self, image_id: str) -> Image:
        request = self._session.get(url_join(self._service_host, Image.resource), params={"search": f"{self.name},{image_id}"})
        if request.status_code == 200:
            return Image(session=self._session, host=self._service_host, **request.json())
        elif request.status_code == 404:
            return None
        else:
            raise RuntimeError("Unable to fetch image")

    def get_cross_validation_file(self, download_to: str = None):
        fold = CrossValidationFolders(self._session, host=self._service_host, dataset=self.name)
        return fold.download(download_to)
